import argparse
import inspect

from .arg import Argument

class Application(object):
    def __init__(self, name=None, usage=None, description=None):
        self.name = name

        self._parser = argparse.ArgumentParser(prog=name,
                                               usage=usage,
                                               description=description)
        self._subparsers = self._parser.add_subparsers(dest='_command')

    def command(self, description=None):
        def register(fn):
            argspec = inspect.getfullargspec(fn)
            parser = self._subparsers.add_parser(fn.__name__)
            parser.set_defaults(_impl=fn)

            for arg_name, arg_annotation in read_annotations(fn):
                if arg_annotation.is_option:
                    arg_kwargs = dict(arg_annotation.kwargs, dest=arg_name)
                else:
                    assert arg_name == arg_annotation.names[0]
                    arg_kwargs = arg_annotation.kwargs

                parser.add_argument(*arg_annotation.names, **arg_kwargs)

            return fn

        return register

    def parse(self, argv=None):
        return self._parser.parse_args(argv)

    def run(self, argv=None):
        args = self.parse(argv)
        fn = args._impl

        kwargs = {arg_name: getattr(args, arg_name)
                  for arg_name, _ in read_annotations(fn)}

        return fn(**kwargs)

    def __repr__(self):
        return '<%s%s>' % (type(self).__name__,
                           ' %s' % self.name if self.name else '')

def read_annotations(fn):
    argspec = inspect.getfullargspec(fn)
    args = []

    for arg in argspec.args:
        ann = argspec.annotations.get(arg)

        if isinstance(ann, Argument):
            args.append((arg, ann))

    return args



import shlex

import pytest

from clinch import application, arg

git = application('git', description='a stupid content tracker')

@git.command('Add things')
def add(all: arg('-A', '--all', action='store_true',
                 help='really make things match'),
        paths: arg('paths', nargs='*')):
    return (add, all, paths)

@git.command('Commit changes')
def commit(all: arg('-a', '--all', action='store_true',
                    help='include all modifications and deletions'),
           message: arg('-m', '--message', help='the commit message'),
           verbose: arg('-v', '--verbose', action='store_true',
                        help='show diff in message editor')):
    return (commit, all, message, verbose)

def test_app():
    assert git.name == 'git'

def test_run():
    def run(cmdline):
        return git.run(shlex.split(cmdline))

    assert run('add') == (add, False, [])
    assert run('add -A') == (add, True, [])
    assert run('add foo bar') == (add, False, ['foo', 'bar'])

    assert run('commit') == (commit, False, None, False)
    assert run('commit -a') == (commit, True, None, False)
    assert run('commit -m "hey there"') == (commit, False, 'hey there', False)
    assert run('commit -av') == (commit, True, None, True)

    with pytest.raises(SystemExit):
        run('lol')

def test_run_sys_argv(monkeypatch):
    monkeypatch.setattr('sys.argv', ['git', 'commit', '-a'])
    assert git.run() == (commit, True, None, False)

def test_repr():
    assert repr(git) == '<Application git>'

if __name__ == '__main__':
    git.run()



class Argument(object):
    def __init__(self, *names, **kwargs):
        assert len(names) > 0
        assert 'dest' not in kwargs, \
            'dest is implicitly set to the argument name'

        self.names = names
        self.kwargs = kwargs

    @property
    def is_option(self):
        return any(name.startswith('-') for name in self.names)

    def __repr__(self):
        parts = (
            ['%s' % name for name in self.names] +
            ['%s=%r' % (k, v) for k, v in self.kwargs.iteritems()]
        )

        return '%s(%s)' % (type(self).__name__, ', '.join(parts))



#!/usr/bin/env python3

import os.path
import sys

from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

root = os.path.abspath(os.path.dirname(__file__))

def read_requirements(filename):
    path = os.path.join(root, filename)
    with open(path, 'rt') as f:
        return [line.strip() for line in f]

class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest
        exit_status = pytest.main(self.pytest_args)
        sys.exit(exit_status)

setup(
    name='clinch',
    version='0.1.0',
    packages=find_packages(),

    author='Eric Naeseth',
    author_email='eric@naeseth.com',

    tests_require=read_requirements('dev-requirements.txt'),
    cmdclass={'test': PyTest},

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
)
